# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""PwPtPythonPane class."""

import asyncio
import functools
import io
import logging
import os
import sys
import shlex
import subprocess
from typing import Iterable, Optional, TYPE_CHECKING
from unittest.mock import patch

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.completion import merge_completers
from prompt_toolkit.filters import (
    Condition,
    has_focus,
    to_filter,
)
from ptpython.completer import (  # type: ignore
    CompletePrivateAttributes,
    PythonCompleter,
)
import ptpython.repl  # type: ignore
from ptpython.layout import (  # type: ignore
    CompletionVisualisation,
    Dimension,
)
import pygments.plugin

from pw_console.pigweed_code_style import (
    PigweedCodeStyle,
    PigweedCodeLightStyle,
)
from pw_console.text_formatting import remove_formatting

if TYPE_CHECKING:
    from pw_console.repl_pane import ReplPane

_LOG = logging.getLogger(__package__)
_SYSTEM_COMMAND_LOG = logging.getLogger('pw_console_system_command')

_original_find_plugin_styles = pygments.plugin.find_plugin_styles


def _wrapped_find_plugin_styles():
    """Patch pygment find_plugin_styles to also include Pigweed codes styles

    This allows using these themes without requiring Python entrypoints.
    """
    for style in [
        ('pigweed-code', PigweedCodeStyle),
        ('pigweed-code-light', PigweedCodeLightStyle),
    ]:
        yield style
    yield from _original_find_plugin_styles()


class MissingPtpythonBufferControl(Exception):
    """Exception for a missing ptpython BufferControl object."""


def _user_input_is_a_shell_command(text: str) -> bool:
    return text.startswith('!')


class PwPtPythonRepl(
    ptpython.repl.PythonRepl
):  # pylint: disable=too-many-instance-attributes
    """A ptpython repl class with changes to code execution and output related
    methods."""

    @patch(
        'pygments.styles.find_plugin_styles', new=_wrapped_find_plugin_styles
    )
    def __init__(
        self,
        *args,
        # pw_console specific kwargs
        extra_completers: Optional[Iterable] = None,
        **ptpython_kwargs,
    ):
        completer = None
        if extra_completers:
            # Create the default python completer used by
            # ptpython.repl.PythonRepl
            python_completer = PythonCompleter(
                # No self.get_globals yet so this must be a lambda
                # pylint: disable=unnecessary-lambda
                lambda: self.get_globals(),
                lambda: self.get_locals(),
                lambda: self.enable_dictionary_completion,  # type: ignore
            )

            all_completers = [python_completer]
            all_completers.extend(extra_completers)
            # Merge default Python completer with the new custom one.
            completer = merge_completers(all_completers)

        super().__init__(
            *args,
            create_app=False,
            # Absolute minimum height of 1
            _input_buffer_height=Dimension(min=1),
            _completer=completer,
            **ptpython_kwargs,
        )

        self.enable_mouse_support: bool = True
        self.enable_history_search: bool = True
        self.enable_dictionary_completion: bool = True
        self._set_pt_python_input_buffer_control_focusable()

        # Change some ptpython.repl defaults.
        self.show_status_bar = False
        self.show_exit_confirmation = False
        self.complete_private_attributes = (
            CompletePrivateAttributes.IF_NO_PUBLIC
        )

        # Function signature that shows args, kwargs, and types under the cursor
        # of the input window.
        self.show_signature: bool = True
        # Docstring of the current completed function that appears at the bottom
        # of the input window.
        self.show_docstring: bool = False

        # Turn off the completion menu in ptpython. The CompletionsMenu in
        # ConsoleApp.root_container will handle this.
        self.completion_visualisation: CompletionVisualisation = (
            CompletionVisualisation.NONE
        )

        # Additional state variables.
        self.repl_pane: 'Optional[ReplPane]' = None
        self._last_result = None
        self._last_exception = None

    def _set_pt_python_input_buffer_control_focusable(self) -> None:
        """Enable focus_on_click for ptpython's input buffer."""
        error_message = (
            'Unable to find ptpythons BufferControl input object.\n'
            '  For the last known position see:\n'
            '  https://github.com/prompt-toolkit/ptpython/'
            'blob/6072174eace5b645b0cfd5b21b4c237e2539f577/'
            'ptpython/layout.py#L598\n'
            '\n'
            'The installed version of ptpython may not be compatible with'
            ' pw console; please try re-running environment setup.'
        )

        try:
            # Fetch the Window's BufferControl object.
            # From ptpython/layout.py:
            #   self.root_container = HSplit([
            #     VSplit([
            #       HSplit([
            #         FloatContainer(
            #           content=HSplit(
            #             [create_python_input_window()] + extra_body
            #           ), ...
            ptpython_buffer_control = (
                self.ptpython_layout.root_container.children[0]  # type: ignore
                .children[0]
                .children[0]
                .content.children[0]
                .content
            )
            # This should be a BufferControl instance
            if not isinstance(ptpython_buffer_control, BufferControl):
                raise MissingPtpythonBufferControl(error_message)
            # Enable focus options
            ptpython_buffer_control.focusable = to_filter(True)
            ptpython_buffer_control.focus_on_click = to_filter(True)
        except IndexError as _error:
            raise MissingPtpythonBufferControl(error_message)

    def __pt_container__(self):
        """Return the prompt_toolkit root container for class.

        This allows self to be used wherever prompt_toolkit expects a container
        object."""
        return self.ptpython_layout.root_container

    def set_repl_pane(self, repl_pane):
        """Update the parent pw_console.ReplPane reference."""
        self.repl_pane = repl_pane

    def _save_result(self, formatted_text):
        """Save the last repl execution result."""
        unformatted_result = remove_formatting(formatted_text)
        self._last_result = unformatted_result

    def _save_exception(self, formatted_text):
        """Save the last repl exception."""
        unformatted_result = remove_formatting(formatted_text)
        self._last_exception = unformatted_result

    def clear_last_result(self):
        """Erase the last repl execution result."""
        self._last_result = None
        self._last_exception = None

    def show_result(self, result):
        """Format and save output results.

        This function is called from the _run_user_code() function which is
        always run from the user code thread, within
        .run_and_show_expression_async().
        """
        formatted_result = self._format_result_output(result)
        self._save_result(formatted_result)

    def _handle_exception(self, e: BaseException) -> None:
        """Format and save output results.

        This function is called from the _run_user_code() function which is
        always run from the user code thread, within
        .run_and_show_expression_async().
        """
        formatted_result = self._format_exception_output(e)
        self._save_exception(formatted_result.__pt_formatted_text__())

    def user_code_complete_callback(self, input_text, future):
        """Callback to run after user repl code is finished."""
        # If there was an exception it will be saved in self._last_result
        result_text = self._last_result
        result_object = None
        exception_text = self._last_exception

        # _last_results consumed, erase for the next run.
        self.clear_last_result()

        stdout_contents = None
        stderr_contents = None
        if future.result():
            future_result = future.result()
            stdout_contents = future_result['stdout']
            stderr_contents = future_result['stderr']
            result_object = future_result['result']

            if result_object is not None:
                # Use ptpython formatted results:
                formatted_result = self._format_result_output(result_object)
                result_text = remove_formatting(formatted_result)

        # Job is finished, append the last result.
        self.repl_pane.append_result_to_executed_code(
            input_text,
            future,
            result_text,
            stdout_contents,
            stderr_contents,
            exception_text=exception_text,
            result_object=result_object,
        )

        # Rebuild output buffer.
        self.repl_pane.update_output_buffer(
            'pw_ptpython_repl.user_code_complete_callback'
        )

        # Trigger a prompt_toolkit application redraw.
        self.repl_pane.application.application.invalidate()

    async def _run_system_command(  # pylint: disable=no-self-use
        self, text, stdout_proxy, _stdin_proxy
    ) -> int:
        """Run a shell command and print results to the repl."""
        command = shlex.split(text)
        returncode = None
        env = os.environ.copy()
        # Force colors in Pigweed subcommands and some terminal apps.
        env['PW_USE_COLOR'] = '1'
        env['CLICOLOR_FORCE'] = '1'

        def _handle_output(output):
            # Force tab characters to 8 spaces to prevent \t from showing in
            # prompt_toolkit.
            output = output.replace('\t', '        ')
            # Strip some ANSI sequences that don't render.
            output = output.replace('\x1b(B\x1b[m', '')
            output = output.replace('\x1b[1m', '')
            stdout_proxy.write(output)
            _SYSTEM_COMMAND_LOG.info(output.rstrip())

        with subprocess.Popen(
            command,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            errors='replace',
        ) as proc:
            # Print the command
            _SYSTEM_COMMAND_LOG.info('')
            _SYSTEM_COMMAND_LOG.info('$ %s', text)
            while returncode is None:
                if not proc.stdout:
                    continue

                # Check for one line and update.
                output = proc.stdout.readline()
                _handle_output(output)

                returncode = proc.poll()

            # Print any remaining lines.
            for output in proc.stdout.readlines():
                _handle_output(output)

        return returncode

    async def _run_user_code(self, text, stdout_proxy, stdin_proxy):
        """Run user code and capture stdout+err.

        This fuction should be run in a separate thread from the main
        prompt_toolkit application."""
        # NOTE: This function runs in a separate thread using the asyncio event
        # loop defined by self.repl_pane.application.user_code_loop. Patching
        # stdout here will not effect the stdout used by prompt_toolkit and the
        # main user interface.

        # Patch stdout and stderr to capture repl print() statements.
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        sys.stdout = stdout_proxy
        sys.stderr = stdin_proxy

        # Run user repl code
        try:
            if _user_input_is_a_shell_command(text):
                result = await self._run_system_command(
                    text[1:], stdout_proxy, stdin_proxy
                )
            else:
                result = await self.run_and_show_expression_async(text)
        finally:
            # Always restore original stdout and stderr
            sys.stdout = original_stdout
            sys.stderr = original_stderr

        # Save the captured output
        stdout_contents = stdout_proxy.getvalue()
        stderr_contents = stdin_proxy.getvalue()

        return {
            'stdout': stdout_contents,
            'stderr': stderr_contents,
            'result': result,
        }

    def _accept_handler(self, buff: Buffer) -> bool:
        """Function executed when pressing enter in the ptpython.repl.PythonRepl
        input buffer."""
        # Do nothing if no text is entered.
        if len(buff.text) == 0:
            return False
        if self.repl_pane is None:
            return False

        repl_input_text = buff.text
        # Exit if quit or exit
        if repl_input_text.strip() in ['quit', 'quit()', 'exit', 'exit()']:
            self.repl_pane.application.application.exit()  # type: ignore

        # Create stdout and stderr proxies
        temp_stdout = io.StringIO()
        temp_stderr = io.StringIO()

        # The help() command with no args uses it's own interactive prompt which
        # will not work if prompt_toolkit is running.
        if repl_input_text.strip() in ['help()']:
            # Run nothing
            repl_input_text = ''
            # Override stdout
            temp_stdout.write(
                'Error: Interactive help() is not compatible with this repl.'
            )

        # Pop open the system command log pane for shell commands.
        if _user_input_is_a_shell_command(repl_input_text):
            self.repl_pane.application.setup_command_runner_log_pane()

        # Execute the repl code in the the separate user_code thread loop.
        future = asyncio.run_coroutine_threadsafe(
            # This function will be executed in a separate thread.
            self._run_user_code(repl_input_text, temp_stdout, temp_stderr),
            # Using this asyncio event loop.
            self.repl_pane.application.user_code_loop,
        )  # type: ignore

        # Save the input text and future object.
        self.repl_pane.append_executed_code(
            repl_input_text, future, temp_stdout, temp_stderr
        )  # type: ignore

        # Run user_code_complete_callback() when done.
        done_callback = functools.partial(
            self.user_code_complete_callback, repl_input_text
        )
        future.add_done_callback(done_callback)

        # Rebuild the parent ReplPane output buffer.
        self.repl_pane.update_output_buffer('pw_ptpython_repl._accept_handler')

        # TODO(tonymd): Return True if exception is found?
        # Don't keep input for now. Return True to keep input text.
        return False

    def line_break_count(self) -> int:
        return self.default_buffer.text.count('\n')

    def input_empty_if_in_focus_condition(self) -> Condition:
        @Condition
        def test() -> bool:
            if has_focus(self)() and len(self.default_buffer.text) == 0:
                return True
            return not has_focus(self)()

        return test



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Console for interacting with devices using HDLC.

To start the console, provide a serial port as the --device argument and paths
or globs for .proto files that define the RPC services to support:

  python -m pw_system.console --device /dev/ttyUSB0 --proto-globs pw_rpc/echo.proto

This starts an IPython console for communicating with the connected device. A
few variables are predefined in the interactive console. These include:

    rpcs   - used to invoke RPCs
    device - the serial device used for communication
    client - the pw_rpc.Client
    protos - protocol buffer messages indexed by proto package

An example echo RPC command:

  rpcs.pw.rpc.EchoService.Echo(msg="hello!")
"""  # pylint: disable=line-too-long

import argparse
import datetime
import glob
from inspect import cleandoc
import logging
from pathlib import Path
import sys
from types import ModuleType
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
)
import socket

import serial
import IPython  # type: ignore

from pw_cli import log as pw_cli_log
from pw_console.embed import PwConsoleEmbed
from pw_console.log_store import LogStore
from pw_console.plugins.bandwidth_toolbar import BandwidthToolbar
from pw_console.pyserial_wrapper import SerialWithLogging
from pw_console.python_logging import create_temp_log_file, JsonLogFormatter
from pw_rpc.console_tools.console import flattened_rpc_completions
from pw_system.device import Device
from pw_tokenizer.detokenize import AutoUpdatingDetokenizer

# Default proto imports:
from pw_log.proto import log_pb2
from pw_metric_proto import metric_service_pb2
from pw_thread_protos import thread_snapshot_service_pb2
from pw_unit_test_proto import unit_test_pb2

_LOG = logging.getLogger('tools')
_DEVICE_LOG = logging.getLogger('rpc_device')
_SERIAL_DEBUG = logging.getLogger('pw_console.serial_debug_logger')
_ROOT_LOG = logging.getLogger()

PW_RPC_MAX_PACKET_SIZE = 256
SOCKET_SERVER = 'localhost'
SOCKET_PORT = 33000
MKFIFO_MODE = 0o666


def _parse_args():
    """Parses and returns the command line arguments."""
    parser = argparse.ArgumentParser(
        prog="python -m pw_system.console", description=__doc__
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--device', help='the serial port to use')
    parser.add_argument(
        '-b',
        '--baudrate',
        type=int,
        default=115200,
        help='the baud rate to use',
    )
    parser.add_argument(
        '--serial-debug',
        action='store_true',
        help=(
            'Enable debug log tracing of all data passed through'
            'pyserial read and write.'
        ),
    )
    parser.add_argument(
        '-o',
        '--output',
        type=argparse.FileType('wb'),
        default=sys.stdout.buffer,
        help=(
            'The file to which to write device output (HDLC channel 1); '
            'provide - or omit for stdout.'
        ),
    )

    # Log file options
    parser.add_argument(
        '--logfile',
        default='pw_console-logs.txt',
        help=(
            'Default log file. This will contain host side '
            'log messages only unles the '
            '--merge-device-and-host-logs argument is used.'
        ),
    )

    parser.add_argument(
        '--merge-device-and-host-logs',
        action='store_true',
        help=(
            'Include device logs in the default --logfile.'
            'These are normally shown in a separate device '
            'only log file.'
        ),
    )

    parser.add_argument(
        '--host-logfile',
        help=(
            'Additional host only log file. Normally all logs in the '
            'default logfile are host only.'
        ),
    )

    parser.add_argument(
        '--device-logfile',
        default='pw_console-device-logs.txt',
        help='Device only log file.',
    )

    parser.add_argument(
        '--json-logfile', help='Device only JSON formatted log file.'
    )

    group.add_argument(
        '-s',
        '--socket-addr',
        type=str,
        help='use socket to connect to server, type default for\
            localhost:33000, or manually input the server address:port',
    )
    parser.add_argument(
        "--token-databases",
        metavar='elf_or_token_database',
        nargs="+",
        type=Path,
        help="Path to tokenizer database csv file(s).",
    )
    parser.add_argument(
        '--config-file',
        type=Path,
        help='Path to a pw_console yaml config file.',
    )
    parser.add_argument(
        '--proto-globs',
        nargs='+',
        default=[],
        help='glob pattern for .proto files.',
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Enables debug logging when set.',
    )
    parser.add_argument(
        '--ipython',
        action='store_true',
        dest='use_ipython',
        help='Use IPython instead of pw_console.',
    )

    # TODO(b/248257406) Use argparse.BooleanOptionalAction when Python 3.8 is
    # no longer supported.
    parser.add_argument(
        '--rpc-logging',
        action='store_true',
        default=True,
        help='Use pw_rpc based logging.',
    )

    parser.add_argument(
        '--no-rpc-logging',
        action='store_false',
        dest='rpc_logging',
        help="Don't use pw_rpc based logging.",
    )

    return parser.parse_args()


def _expand_globs(globs: Iterable[str]) -> Iterator[Path]:
    for pattern in globs:
        for file in glob.glob(pattern, recursive=True):
            yield Path(file)


def _start_python_terminal(  # pylint: disable=too-many-arguments
    device: Device,
    device_log_store: LogStore,
    root_log_store: LogStore,
    serial_debug_log_store: LogStore,
    log_file: str,
    host_logfile: str,
    device_logfile: str,
    json_logfile: str,
    serial_debug: bool = False,
    config_file_path: Optional[Path] = None,
    use_ipython: bool = False,
) -> None:
    """Starts an interactive Python terminal with preset variables."""
    local_variables = dict(
        client=device.client,
        device=device,
        rpcs=device.rpcs,
        protos=device.client.protos.packages,
        # Include the active pane logger for creating logs in the repl.
        DEVICE_LOG=_DEVICE_LOG,
        LOG=logging.getLogger(),
    )

    welcome_message = cleandoc(
        """
        Welcome to the Pigweed Console!

        Help: Press F1 or click the [Help] menu
        To move focus: Press Shift-Tab or click on a window

        Example Python commands:

          device.rpcs.pw.rpc.EchoService.Echo(msg='hello!')
          LOG.warning('Message appears in Host Logs window.')
          DEVICE_LOG.warning('Message appears in Device Logs window.')
    """
    )

    welcome_message += '\n\nLogs are being saved to:\n  ' + log_file
    if host_logfile:
        welcome_message += '\nHost logs are being saved to:\n  ' + host_logfile
    if device_logfile:
        welcome_message += (
            '\nDevice logs are being saved to:\n  ' + device_logfile
        )
    if json_logfile:
        welcome_message += (
            '\nJSON device logs are being saved to:\n  ' + json_logfile
        )

    if use_ipython:
        print(welcome_message)
        IPython.terminal.embed.InteractiveShellEmbed().mainloop(
            local_ns=local_variables, module=argparse.Namespace()
        )
        return

    client_info = device.info()
    completions = flattened_rpc_completions([client_info])

    log_windows: Dict[str, Union[List[logging.Logger], LogStore]] = {
        'Device Logs': device_log_store,
        'Host Logs': root_log_store,
    }
    if serial_debug:
        log_windows['Serial Debug'] = serial_debug_log_store

    interactive_console = PwConsoleEmbed(
        global_vars=local_variables,
        local_vars=None,
        loggers=log_windows,
        repl_startup_message=welcome_message,
        help_text=__doc__,
        config_file_path=config_file_path,
    )
    interactive_console.add_sentence_completer(completions)
    if serial_debug:
        interactive_console.add_bottom_toolbar(BandwidthToolbar())

    # Setup Python logger propagation
    interactive_console.setup_python_logging(
        # Send any unhandled log messages to the external file.
        last_resort_filename=log_file,
        # Don't change propagation for these loggers.
        loggers_with_no_propagation=[_DEVICE_LOG],
    )

    interactive_console.embed()


class SocketClientImpl:
    def __init__(self, config: str):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_server = ''
        socket_port = 0

        if config == 'default':
            socket_server = SOCKET_SERVER
            socket_port = SOCKET_PORT
        else:
            socket_server, socket_port_str = config.split(':')
            socket_port = int(socket_port_str)
        self.socket.connect((socket_server, socket_port))

    def write(self, data: bytes):
        self.socket.sendall(data)

    def read(self, num_bytes: int = PW_RPC_MAX_PACKET_SIZE):
        return self.socket.recv(num_bytes)


# pylint: disable=too-many-arguments,too-many-locals
def console(
    device: str,
    baudrate: int,
    proto_globs: Collection[str],
    token_databases: Collection[Path],
    socket_addr: str,
    logfile: str,
    host_logfile: str,
    device_logfile: str,
    json_logfile: str,
    output: Any,
    serial_debug: bool = False,
    config_file: Optional[Path] = None,
    verbose: bool = False,
    compiled_protos: Optional[List[ModuleType]] = None,
    merge_device_and_host_logs: bool = False,
    rpc_logging: bool = True,
    use_ipython: bool = False,
) -> int:
    """Starts an interactive RPC console for HDLC."""
    # argparse.FileType doesn't correctly handle '-' for binary files.
    if output is sys.stdout:
        output = sys.stdout.buffer

    # Don't send device logs to the root logger.
    _DEVICE_LOG.propagate = False
    # Create pw_console LogStore handlers. These are the data source for log
    # messages to be displayed in the UI.
    device_log_store = LogStore()
    root_log_store = LogStore()
    serial_debug_log_store = LogStore()
    # Attach the LogStores as handlers for each log window we want to show.
    # This should be done before device initialization to capture early
    # messages.
    _DEVICE_LOG.addHandler(device_log_store)
    _ROOT_LOG.addHandler(root_log_store)
    _SERIAL_DEBUG.addHandler(serial_debug_log_store)

    if not logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        logfile = create_temp_log_file()

    log_level = logging.DEBUG if verbose else logging.INFO

    pw_cli_log.install(
        level=log_level, use_color=False, hide_timestamp=False, log_file=logfile
    )

    if device_logfile:
        pw_cli_log.install(
            level=log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=device_logfile,
            logger=_DEVICE_LOG,
        )
    if host_logfile:
        pw_cli_log.install(
            level=log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=host_logfile,
            logger=_ROOT_LOG,
        )

    if merge_device_and_host_logs:
        # Add device logs to the default logfile.
        pw_cli_log.install(
            level=log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=logfile,
            logger=_DEVICE_LOG,
        )

    _LOG.setLevel(log_level)
    _DEVICE_LOG.setLevel(log_level)
    _ROOT_LOG.setLevel(log_level)
    _SERIAL_DEBUG.setLevel(logging.DEBUG)

    if json_logfile:
        json_filehandler = logging.FileHandler(json_logfile, encoding='utf-8')
        json_filehandler.setLevel(log_level)
        json_filehandler.setFormatter(JsonLogFormatter())
        _DEVICE_LOG.addHandler(json_filehandler)

    detokenizer = None
    if token_databases:
        detokenizer = AutoUpdatingDetokenizer(*token_databases)
        detokenizer.show_errors = True

    protos: List[Union[ModuleType, Path]] = list(_expand_globs(proto_globs))

    if compiled_protos is None:
        compiled_protos = []

    # Append compiled log.proto library to avoid include errors when manually
    # provided, and shadowing errors due to ordering when the default global
    # search path is used.
    if rpc_logging:
        compiled_protos.append(log_pb2)
    compiled_protos.append(unit_test_pb2)
    protos.extend(compiled_protos)
    protos.append(metric_service_pb2)
    protos.append(thread_snapshot_service_pb2)

    if not protos:
        _LOG.critical(
            'No .proto files were found with %s', ', '.join(proto_globs)
        )
        _LOG.critical('At least one .proto file is required')
        return 1

    _LOG.debug(
        'Found %d .proto files found with %s',
        len(protos),
        ', '.join(proto_globs),
    )

    serial_impl = serial.Serial
    if serial_debug:
        serial_impl = SerialWithLogging

    timestamp_decoder = None
    if socket_addr is None:
        serial_device = serial_impl(
            device,
            baudrate,
            # Timeout in seconds. This should be a very small value. Setting to
            # zero makes pyserial read() non-blocking which will cause the host
            # machine to busy loop and 100% CPU usage.
            # https://pythonhosted.org/pyserial/pyserial_api.html#serial.Serial
            timeout=0.1,
        )
        read = lambda: serial_device.read(8192)
        write = serial_device.write

        # Overwrite decoder for serial device.
        def milliseconds_to_string(timestamp):
            """Parses milliseconds since boot to a human-readable string."""
            return str(datetime.timedelta(seconds=timestamp / 1e3))[:-3]

        timestamp_decoder = milliseconds_to_string
    else:
        try:
            socket_device = SocketClientImpl(socket_addr)
            read = socket_device.read
            write = socket_device.write
        except ValueError:
            _LOG.exception('Failed to initialize socket at %s', socket_addr)
            return 1

    device_client = Device(
        1,
        read,
        write,
        protos,
        detokenizer,
        timestamp_decoder=timestamp_decoder,
        rpc_timeout_s=5,
        use_rpc_logging=rpc_logging,
    )

    _start_python_terminal(
        device=device_client,
        device_log_store=device_log_store,
        root_log_store=root_log_store,
        serial_debug_log_store=serial_debug_log_store,
        log_file=logfile,
        host_logfile=host_logfile,
        device_logfile=device_logfile,
        json_logfile=json_logfile,
        serial_debug=serial_debug,
        config_file_path=config_file,
        use_ipython=use_ipython,
    )
    return 0


def main() -> int:
    return console(**vars(_parse_args()))


def main_with_compiled_protos(compiled_protos):
    return console(**vars(_parse_args()), compiled_protos=compiled_protos)


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_transfer/transfer.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_transfer/transfer.proto',
  package='pw.transfer',
  syntax='proto3',
  serialized_options=b'\n\027dev.pigweed.pw_transferP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1apw_transfer/transfer.proto\x12\x0bpw.transfer\"\xad\x05\n\x05\x43hunk\x12\x13\n\x0btransfer_id\x18\x01 \x01(\r\x12\x1a\n\rpending_bytes\x18\x02 \x01(\rH\x00\x88\x01\x01\x12!\n\x14max_chunk_size_bytes\x18\x03 \x01(\rH\x01\x88\x01\x01\x12#\n\x16min_delay_microseconds\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x0e\n\x06offset\x18\x05 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x1c\n\x0fremaining_bytes\x18\x07 \x01(\x04H\x03\x88\x01\x01\x12\x13\n\x06status\x18\x08 \x01(\rH\x04\x88\x01\x01\x12\x19\n\x11window_end_offset\x18\t \x01(\r\x12*\n\x04type\x18\n \x01(\x0e\x32\x17.pw.transfer.Chunk.TypeH\x05\x88\x01\x01\x12\x18\n\x0bresource_id\x18\x0b \x01(\rH\x06\x88\x01\x01\x12\x17\n\nsession_id\x18\x0c \x01(\rH\x07\x88\x01\x01\x12\x1d\n\x10protocol_version\x18\r \x01(\rH\x08\x88\x01\x01\"\x9e\x01\n\x04Type\x12\x08\n\x04\x44\x41TA\x10\x00\x12\t\n\x05START\x10\x01\x12\x19\n\x15PARAMETERS_RETRANSMIT\x10\x02\x12\x17\n\x13PARAMETERS_CONTINUE\x10\x03\x12\x0e\n\nCOMPLETION\x10\x04\x12\x12\n\x0e\x43OMPLETION_ACK\x10\x05\x12\r\n\tSTART_ACK\x10\x06\x12\x1a\n\x16START_ACK_CONFIRMATION\x10\x07\x42\x10\n\x0e_pending_bytesB\x17\n\x15_max_chunk_size_bytesB\x19\n\x17_min_delay_microsecondsB\x12\n\x10_remaining_bytesB\t\n\x07_statusB\x07\n\x05_typeB\x0e\n\x0c_resource_idB\r\n\x0b_session_idB\x13\n\x11_protocol_version2s\n\x08Transfer\x12\x32\n\x04Read\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x12\x33\n\x05Write\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x42\x1b\n\x17\x64\x65v.pigweed.pw_transferP\x01\x62\x06proto3'
)



_CHUNK_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='pw.transfer.Chunk.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DATA', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='START', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PARAMETERS_RETRANSMIT', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PARAMETERS_CONTINUE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPLETION', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPLETION_ACK', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='START_ACK', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='START_ACK_CONFIRMATION', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=409,
  serialized_end=567,
)
_sym_db.RegisterEnumDescriptor(_CHUNK_TYPE)


_CHUNK = _descriptor.Descriptor(
  name='Chunk',
  full_name='pw.transfer.Chunk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transfer_id', full_name='pw.transfer.Chunk.transfer_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pending_bytes', full_name='pw.transfer.Chunk.pending_bytes', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_chunk_size_bytes', full_name='pw.transfer.Chunk.max_chunk_size_bytes', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_delay_microseconds', full_name='pw.transfer.Chunk.min_delay_microseconds', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='pw.transfer.Chunk.offset', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='pw.transfer.Chunk.data', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remaining_bytes', full_name='pw.transfer.Chunk.remaining_bytes', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='pw.transfer.Chunk.status', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='window_end_offset', full_name='pw.transfer.Chunk.window_end_offset', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='pw.transfer.Chunk.type', index=9,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_id', full_name='pw.transfer.Chunk.resource_id', index=10,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='pw.transfer.Chunk.session_id', index=11,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='protocol_version', full_name='pw.transfer.Chunk.protocol_version', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CHUNK_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_pending_bytes', full_name='pw.transfer.Chunk._pending_bytes',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_max_chunk_size_bytes', full_name='pw.transfer.Chunk._max_chunk_size_bytes',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_min_delay_microseconds', full_name='pw.transfer.Chunk._min_delay_microseconds',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_remaining_bytes', full_name='pw.transfer.Chunk._remaining_bytes',
      index=3, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_status', full_name='pw.transfer.Chunk._status',
      index=4, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_type', full_name='pw.transfer.Chunk._type',
      index=5, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_resource_id', full_name='pw.transfer.Chunk._resource_id',
      index=6, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_session_id', full_name='pw.transfer.Chunk._session_id',
      index=7, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_protocol_version', full_name='pw.transfer.Chunk._protocol_version',
      index=8, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=44,
  serialized_end=729,
)

_CHUNK.fields_by_name['type'].enum_type = _CHUNK_TYPE
_CHUNK_TYPE.containing_type = _CHUNK
_CHUNK.oneofs_by_name['_pending_bytes'].fields.append(
  _CHUNK.fields_by_name['pending_bytes'])
_CHUNK.fields_by_name['pending_bytes'].containing_oneof = _CHUNK.oneofs_by_name['_pending_bytes']
_CHUNK.oneofs_by_name['_max_chunk_size_bytes'].fields.append(
  _CHUNK.fields_by_name['max_chunk_size_bytes'])
_CHUNK.fields_by_name['max_chunk_size_bytes'].containing_oneof = _CHUNK.oneofs_by_name['_max_chunk_size_bytes']
_CHUNK.oneofs_by_name['_min_delay_microseconds'].fields.append(
  _CHUNK.fields_by_name['min_delay_microseconds'])
_CHUNK.fields_by_name['min_delay_microseconds'].containing_oneof = _CHUNK.oneofs_by_name['_min_delay_microseconds']
_CHUNK.oneofs_by_name['_remaining_bytes'].fields.append(
  _CHUNK.fields_by_name['remaining_bytes'])
_CHUNK.fields_by_name['remaining_bytes'].containing_oneof = _CHUNK.oneofs_by_name['_remaining_bytes']
_CHUNK.oneofs_by_name['_status'].fields.append(
  _CHUNK.fields_by_name['status'])
_CHUNK.fields_by_name['status'].containing_oneof = _CHUNK.oneofs_by_name['_status']
_CHUNK.oneofs_by_name['_type'].fields.append(
  _CHUNK.fields_by_name['type'])
_CHUNK.fields_by_name['type'].containing_oneof = _CHUNK.oneofs_by_name['_type']
_CHUNK.oneofs_by_name['_resource_id'].fields.append(
  _CHUNK.fields_by_name['resource_id'])
_CHUNK.fields_by_name['resource_id'].containing_oneof = _CHUNK.oneofs_by_name['_resource_id']
_CHUNK.oneofs_by_name['_session_id'].fields.append(
  _CHUNK.fields_by_name['session_id'])
_CHUNK.fields_by_name['session_id'].containing_oneof = _CHUNK.oneofs_by_name['_session_id']
_CHUNK.oneofs_by_name['_protocol_version'].fields.append(
  _CHUNK.fields_by_name['protocol_version'])
_CHUNK.fields_by_name['protocol_version'].containing_oneof = _CHUNK.oneofs_by_name['_protocol_version']
DESCRIPTOR.message_types_by_name['Chunk'] = _CHUNK
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Chunk = _reflection.GeneratedProtocolMessageType('Chunk', (_message.Message,), {
  'DESCRIPTOR' : _CHUNK,
  '__module__' : 'pw_transfer.transfer_pb2'
  # @@protoc_insertion_point(class_scope:pw.transfer.Chunk)
  })
_sym_db.RegisterMessage(Chunk)


DESCRIPTOR._options = None

_TRANSFER = _descriptor.ServiceDescriptor(
  name='Transfer',
  full_name='pw.transfer.Transfer',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=731,
  serialized_end=846,
  methods=[
  _descriptor.MethodDescriptor(
    name='Read',
    full_name='pw.transfer.Transfer.Read',
    index=0,
    containing_service=None,
    input_type=_CHUNK,
    output_type=_CHUNK,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Write',
    full_name='pw.transfer.Transfer.Write',
    index=1,
    containing_service=None,
    input_type=_CHUNK,
    output_type=_CHUNK,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRANSFER)

DESCRIPTOR.services_by_name['Transfer'] = _TRANSFER

# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.

Implementation of metadata formats specified in TUF Specification.
See https://theupdateframework.github.io/specification/latest/
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _HashFunction:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _HashFunctionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_HashFunction.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN_HASH_FUNCTION: _HashFunction.ValueType  # 0
    """Never use this in any TUF metadata."""
    SHA256: _HashFunction.ValueType  # 1

class HashFunction(_HashFunction, metaclass=_HashFunctionEnumTypeWrapper): ...

UNKNOWN_HASH_FUNCTION: HashFunction.ValueType  # 0
"""Never use this in any TUF metadata."""
SHA256: HashFunction.ValueType  # 1
global___HashFunction = HashFunction

class _KeyType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _KeyTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_KeyType.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN_KEY_TYPE: _KeyType.ValueType  # 0
    """Never use this in any TUF metadata."""
    RSA: _KeyType.ValueType  # 1
    ED25519: _KeyType.ValueType  # 2
    ECDSA_SHA2_NISTP256: _KeyType.ValueType  # 3

class KeyType(_KeyType, metaclass=_KeyTypeEnumTypeWrapper): ...

UNKNOWN_KEY_TYPE: KeyType.ValueType  # 0
"""Never use this in any TUF metadata."""
RSA: KeyType.ValueType  # 1
ED25519: KeyType.ValueType  # 2
ECDSA_SHA2_NISTP256: KeyType.ValueType  # 3
global___KeyType = KeyType

class _KeyScheme:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _KeySchemeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_KeyScheme.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN_KEY_SCHEME: _KeyScheme.ValueType  # 0
    """Never use this in any TUF metadata."""
    RSASSA_PSS_SHA256_SCHEME: _KeyScheme.ValueType  # 1
    """RSA Probabilistic signature scheme with appendix.
    The underlying hash function is SHA256.
    In TUF Specification, this is referred to as "rsassa-pss-sha256".
    """
    ED25519_SCHEME: _KeyScheme.ValueType  # 2
    """Elliptic Curve digital signature algorithm based on Twisted Edwards curves.
    See https://ed25519.cr.yp.to/.
    In TUF Specification, it is referred to as "ed25519".
    """
    ECDSA_SHA2_NISTP256_SCHEME: _KeyScheme.ValueType  # 3
    """Elliptic Curve Digital Signature Algorithm with NIST P-256 curve signing
    and SHA-256 hashing. See
    https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm In
    TUF Specification, it is referred to as "ecdsa-sha2-nistp256".
    """

class KeyScheme(_KeyScheme, metaclass=_KeySchemeEnumTypeWrapper): ...

UNKNOWN_KEY_SCHEME: KeyScheme.ValueType  # 0
"""Never use this in any TUF metadata."""
RSASSA_PSS_SHA256_SCHEME: KeyScheme.ValueType  # 1
"""RSA Probabilistic signature scheme with appendix.
The underlying hash function is SHA256.
In TUF Specification, this is referred to as "rsassa-pss-sha256".
"""
ED25519_SCHEME: KeyScheme.ValueType  # 2
"""Elliptic Curve digital signature algorithm based on Twisted Edwards curves.
See https://ed25519.cr.yp.to/.
In TUF Specification, it is referred to as "ed25519".
"""
ECDSA_SHA2_NISTP256_SCHEME: KeyScheme.ValueType  # 3
"""Elliptic Curve Digital Signature Algorithm with NIST P-256 curve signing
and SHA-256 hashing. See
https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm In
TUF Specification, it is referred to as "ecdsa-sha2-nistp256".
"""
global___KeyScheme = KeyScheme

class SignedRootMetadata(google.protobuf.message.Message):
    """Metadata for a particular TUF role (e.g. targets metadata).
    Was TufMetadata
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_ROOT_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    serialized_root_metadata: builtins.bytes
    """Serialized RootMetadata message that is the data portion of the metadata."""
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]:
        """Signature of the canonical form of the role's serialized metadata
        (serialized_root_metadata).
        """
    def __init__(
        self,
        *,
        serialized_root_metadata: builtins.bytes = ...,
        signatures: collections.abc.Iterable[global___Signature] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["serialized_root_metadata", b"serialized_root_metadata", "signatures", b"signatures"]) -> None: ...

global___SignedRootMetadata = SignedRootMetadata

class SignedTimestampMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_TIMESTAMP_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    serialized_timestamp_metadata: builtins.bytes
    """Serialized TimestampMetadata message that is the data portion of the
    metadata.
    """
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]:
        """Signature of the canonical form of the role's serialized metadata
        (serialized_timestamp_metadata).
        """
    def __init__(
        self,
        *,
        serialized_timestamp_metadata: builtins.bytes = ...,
        signatures: collections.abc.Iterable[global___Signature] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["serialized_timestamp_metadata", b"serialized_timestamp_metadata", "signatures", b"signatures"]) -> None: ...

global___SignedTimestampMetadata = SignedTimestampMetadata

class SignedSnapshotMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_SNAPSHOT_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    serialized_snapshot_metadata: builtins.bytes
    """Serialized SnapshotMetadata message that is the data portion of the
    metadata.
    """
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]:
        """Signature of the canonical form of the role's serialized metadata
        (serialized_snapshot_metadata).
        """
    def __init__(
        self,
        *,
        serialized_snapshot_metadata: builtins.bytes = ...,
        signatures: collections.abc.Iterable[global___Signature] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["serialized_snapshot_metadata", b"serialized_snapshot_metadata", "signatures", b"signatures"]) -> None: ...

global___SignedSnapshotMetadata = SignedSnapshotMetadata

class SignedTargetsMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_TARGETS_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    serialized_targets_metadata: builtins.bytes
    """Serialized TargetsMetadata message that is the data portion of the
    metadata.
    """
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]:
        """Signature of the canonical form of the role's serialized metadata
        (serialized_targets_metadata).
        """
    def __init__(
        self,
        *,
        serialized_targets_metadata: builtins.bytes = ...,
        signatures: collections.abc.Iterable[global___Signature] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["serialized_targets_metadata", b"serialized_targets_metadata", "signatures", b"signatures"]) -> None: ...

global___SignedTargetsMetadata = SignedTargetsMetadata

class CommonMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPEC_VERSION_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    EXPIRES_FIELD_NUMBER: builtins.int
    ROLE_FIELD_NUMBER: builtins.int
    spec_version: builtins.str
    """Version number of the TUF Specification.
    Follows the Semantic Versioning 2.0.0 (semver) format. Metadata is
    written according to this version, and clients MUST verify that
    "spec_version" matches the expected version number.
    E.g. "1.0.0".
    """
    version: builtins.int
    """Metadata file version.
    Clients MUST NOT replace a metadata file with a version number less than
    the one currently trusted.
    """
    @property
    def expires(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Expiration time for the metadata.
        Indicates when this metadata should be considered expired and no longer
        trusted by clients. Notice the TUF Specification defines this as a JSON
        string following the ISO 8601 standard. The expected format of the date and
        time string is "YYYY-MM-DDTHH:MM:SSZ". Time is always in UTC, and the "Z"
        time zone designator is attached to indicate a zero UTC offset.
        E.g. "2030-08-26T16:48:27Z".
        """
    role: builtins.str
    """Role type for the metadata.
    Indicates the type of the metadata. Valid values are 'root', 'targets',
    'snapshot' and 'timestamp' as defined in the TUF spec, though we don't
    plan to support 'mirrors'.

    This field serves as a "magic code" that identifies a particular type of
    a metadata. During verification, the client is expected to check this
    field against the expected role type immediately after verifying the
    signatures of a metadata. This can be considered a "confidence booster"
    in the absence of canonical protobuf -- i.e. it makes the various
    `serialized_x_metadata` fields more tamper resistant.
    """
    def __init__(
        self,
        *,
        spec_version: builtins.str = ...,
        version: builtins.int = ...,
        expires: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        role: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_expires", b"_expires", "_role", b"_role", "expires", b"expires", "role", b"role"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_expires", b"_expires", "_role", b"_role", "expires", b"expires", "role", b"role", "spec_version", b"spec_version", "version", b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_expires", b"_expires"]) -> typing_extensions.Literal["expires"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_role", b"_role"]) -> typing_extensions.Literal["role"] | None: ...

global___CommonMetadata = CommonMetadata

class RootMetadata(google.protobuf.message.Message):
    """This content is signed."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMON_METADATA_FIELD_NUMBER: builtins.int
    CONSISTENT_SNAPSHOT_FIELD_NUMBER: builtins.int
    KEYS_FIELD_NUMBER: builtins.int
    ROOT_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    TIMESTAMP_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    SNAPSHOT_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    TARGETS_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    consistent_snapshot: builtins.bool
    """Whether the repo supports consistent snapshots. If the repo has frequent
    updates, you should set this to true.
    """
    @property
    def keys(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyMapping]:
        """Map from Keyid to Key.
        Keyid is a unique identifier that identifies a cryptographic key.
        Contains all of cryptographic keys used by this repository.
        """
    @property
    def root_signature_requirement(self) -> global___SignatureRequirement:
        """KeyConfig is the list of keys use for a particular role and the threshold.
        Threshold is number of keys of that role whose signatures are required in
        order to consider a file as being properly signed by that role.
        """
    @property
    def timestamp_signature_requirement(self) -> global___SignatureRequirement: ...
    @property
    def snapshot_signature_requirement(self) -> global___SignatureRequirement: ...
    @property
    def targets_signature_requirement(self) -> global___SignatureRequirement: ...
    def __init__(
        self,
        *,
        common_metadata: global___CommonMetadata | None = ...,
        consistent_snapshot: builtins.bool = ...,
        keys: collections.abc.Iterable[global___KeyMapping] | None = ...,
        root_signature_requirement: global___SignatureRequirement | None = ...,
        timestamp_signature_requirement: global___SignatureRequirement | None = ...,
        snapshot_signature_requirement: global___SignatureRequirement | None = ...,
        targets_signature_requirement: global___SignatureRequirement | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "root_signature_requirement", b"root_signature_requirement", "snapshot_signature_requirement", b"snapshot_signature_requirement", "targets_signature_requirement", b"targets_signature_requirement", "timestamp_signature_requirement", b"timestamp_signature_requirement"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "consistent_snapshot", b"consistent_snapshot", "keys", b"keys", "root_signature_requirement", b"root_signature_requirement", "snapshot_signature_requirement", b"snapshot_signature_requirement", "targets_signature_requirement", b"targets_signature_requirement", "timestamp_signature_requirement", b"timestamp_signature_requirement"]) -> None: ...

global___RootMetadata = RootMetadata

class TimestampMetadata(google.protobuf.message.Message):
    """The timestamp role is used for freshness check of the snapshot. Any
    project-specific update metadata should go in the top-level targets_metadata
    or with the TargetFile information
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMON_METADATA_FIELD_NUMBER: builtins.int
    SNAPSHOT_METADATA_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    @property
    def snapshot_metadata(self) -> global___MetadataFile:
        """Only one snapshot_metadata is used per timestamp."""
    def __init__(
        self,
        *,
        common_metadata: global___CommonMetadata | None = ...,
        snapshot_metadata: global___MetadataFile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "snapshot_metadata", b"snapshot_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "snapshot_metadata", b"snapshot_metadata"]) -> None: ...

global___TimestampMetadata = TimestampMetadata

class SnapshotMetadata(google.protobuf.message.Message):
    """The snapshot role is used to ensure that the collection of targets_metadata
    files is securely consistent (no target metadata mix and match). Any
    project-specific update metadata should go in the top-level targets_metadata
    or with the TargetFile information
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMON_METADATA_FIELD_NUMBER: builtins.int
    TARGETS_METADATA_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    @property
    def targets_metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MetadataFile]:
        """Map from Target metadata file name to MetadataFile.
        File name can be an arbitrary name or a full file name with relative path.
        This map should contain an entry for the top level targets role and all
        delegated roles.
        """
    def __init__(
        self,
        *,
        common_metadata: global___CommonMetadata | None = ...,
        targets_metadata: collections.abc.Iterable[global___MetadataFile] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "targets_metadata", b"targets_metadata"]) -> None: ...

global___SnapshotMetadata = SnapshotMetadata

class TargetsMetadata(google.protobuf.message.Message):
    """The targets role describes the target files that comprise the software
    update. Targets metadata is organized in to a top-level targets metadata file
    and optional multiple deligated targets metadata files

    The top-level targets metatdata is the correct place to put any
    project-specific build version information, including build ID, hardware rev,
    etc.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMON_METADATA_FIELD_NUMBER: builtins.int
    TARGET_FILES_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    @property
    def target_files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TargetFile]:
        """Collection of target file information
        Target file name can be an arbitrary name or a path that describes where
        the file lives relative to the base directory of the repository, e.g.
        "path/to/amber_tools/0".
        """
    def __init__(
        self,
        *,
        common_metadata: global___CommonMetadata | None = ...,
        target_files: collections.abc.Iterable[global___TargetFile] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["common_metadata", b"common_metadata", "target_files", b"target_files"]) -> None: ...

global___TargetsMetadata = TargetsMetadata

class Signature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_ID_FIELD_NUMBER: builtins.int
    SIG_FIELD_NUMBER: builtins.int
    key_id: builtins.bytes
    """Identifier of the key, which is bytes of the SHA-256 hash of the
    canonical form of the key.
    """
    sig: builtins.bytes
    """The signature of the canonical form of the role's serialized metadata
    (serialized_{root,timestamp,snapshot,targets}_metadata).
    """
    def __init__(
        self,
        *,
        key_id: builtins.bytes = ...,
        sig: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key_id", b"key_id", "sig", b"sig"]) -> None: ...

global___Signature = Signature

class KeyMapping(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_ID_FIELD_NUMBER: builtins.int
    KEY_FIELD_NUMBER: builtins.int
    key_id: builtins.bytes
    """Identifier of the key, which is bytes of the SHA-256 hash of the
    canonical form of the key.
    """
    @property
    def key(self) -> global___Key:
        """Cryptographic key"""
    def __init__(
        self,
        *,
        key_id: builtins.bytes = ...,
        key: global___Key | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["key", b"key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "key_id", b"key_id"]) -> None: ...

global___KeyMapping = KeyMapping

class Key(google.protobuf.message.Message):
    """Identifies an asymmetric cryptographic key."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_TYPE_FIELD_NUMBER: builtins.int
    SCHEME_FIELD_NUMBER: builtins.int
    KEYVAL_FIELD_NUMBER: builtins.int
    key_type: global___KeyType.ValueType
    """Denotes a public key signature system, such as RSA or ECDSA."""
    scheme: global___KeyScheme.ValueType
    """Denotes the signature scheme corresponding to the key type. For example:
    "rsassa-pss-sha256" or "ecdsa-sha2-nistp256".
    """
    keyval: builtins.bytes
    """Stores the serialized public key for this cryptographic algorithm."""
    def __init__(
        self,
        *,
        key_type: global___KeyType.ValueType = ...,
        scheme: global___KeyScheme.ValueType = ...,
        keyval: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key_type", b"key_type", "keyval", b"keyval", "scheme", b"scheme"]) -> None: ...

global___Key = Key

class SignatureRequirement(google.protobuf.message.Message):
    """The set of cryptographic keys used by a specific role. For example, list of
    key_ids used by the top level role "root".
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_IDS_FIELD_NUMBER: builtins.int
    THRESHOLD_FIELD_NUMBER: builtins.int
    @property
    def key_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
        """Set of Keyid's.
        Keyid is a unique identifier that identifies a cryptographic key.
        E.g. "f2d5020d08aea06a0a9192eb6a4f549e17032ebefa1aa9ac167c1e3e727930d6".
        """
    threshold: builtins.int
    """Threshold of signatures required to trust given file.
    In other words; the number of keys of that role whose signatures are
    required in order to consider a file as being properly signed by that role.
    """
    def __init__(
        self,
        *,
        key_ids: collections.abc.Iterable[builtins.bytes] | None = ...,
        threshold: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key_ids", b"key_ids", "threshold", b"threshold"]) -> None: ...

global___SignatureRequirement = SignatureRequirement

class Hash(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FUNCTION_FIELD_NUMBER: builtins.int
    HASH_FIELD_NUMBER: builtins.int
    function: global___HashFunction.ValueType
    hash: builtins.bytes
    """Digest of the cryptographic hash function computed on the target file."""
    def __init__(
        self,
        *,
        function: global___HashFunction.ValueType = ...,
        hash: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["function", b"function", "hash", b"hash"]) -> None: ...

global___Hash = Hash

class TargetFile(google.protobuf.message.Message):
    """Descriptor for a file stored in this repository. Linked to from target
    metadata.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_NAME_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    HASHES_FIELD_NUMBER: builtins.int
    file_name: builtins.str
    """Target file name can be an arbitrary name or a path that describes where
    the file lives relative to the base directory of the repository, e.g.
    "path/to/amber_tools/0".
    """
    length: builtins.int
    """Size of the target file (element payload) in bytes. This the size as stored
    in the bundle. The final applied size can be different due to optional
    compression.
    """
    @property
    def hashes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Hash]:
        """Map from algorithm name to Hash.
        Algorithm name is the name of a cryptographic hash function. E.g. "sha256".
        The Hash string is the hex digest of the cryptographic function computed on
        the target file. E.g.
        "65b8c67f51c993d898250f40aa57a317d854900b3a04895464313e48785440da".
        """
    def __init__(
        self,
        *,
        file_name: builtins.str = ...,
        length: builtins.int = ...,
        hashes: collections.abc.Iterable[global___Hash] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["file_name", b"file_name", "hashes", b"hashes", "length", b"length"]) -> None: ...

global___TargetFile = TargetFile

class MetadataFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    HASHES_FIELD_NUMBER: builtins.int
    file_name: builtins.str
    """Target file name can be an arbitrary name or a path that describes where
    the file lives relative to the base directory of the repository, e.g.
    "path/to/target/0".
    """
    version: builtins.int
    """Metadata file version. E.g. 3."""
    length: builtins.int
    """Size of the target file in bytes."""
    @property
    def hashes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Hash]:
        """Map from algorithm name to Hash.
        Algorithm name is the name of a cryptographic hash function. E.g. "sha256".
        The Hash is the hex digest of the cryptographic function computed on the
        target file. E.g.
        "65b8c67f51c993d898250f40aa57a317d854900b3a04895464313e48785440da".
        """
    def __init__(
        self,
        *,
        file_name: builtins.str | None = ...,
        version: builtins.int = ...,
        length: builtins.int | None = ...,
        hashes: collections.abc.Iterable[global___Hash] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_file_name", b"_file_name", "_length", b"_length", "file_name", b"file_name", "length", b"length"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_file_name", b"_file_name", "_length", b"_length", "file_name", b"file_name", "hashes", b"hashes", "length", b"length", "version", b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_file_name", b"_file_name"]) -> typing_extensions.Literal["file_name"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_length", b"_length"]) -> typing_extensions.Literal["length"] | None: ...

global___MetadataFile = MetadataFile



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapper fuctions to add borders around prompt_toolkit containers."""

from typing import Callable, List, Optional, Union

from prompt_toolkit.layout import (
    AnyContainer,
    FormattedTextControl,
    HSplit,
    VSplit,
    Window,
)


def create_border(
    # pylint: disable=too-many-arguments
    content: AnyContainer,
    content_height: Optional[int] = None,
    title: Union[Callable[[], str], str] = '',
    border_style: Union[Callable[[], str], str] = '',
    base_style: Union[Callable[[], str], str] = '',
    top: bool = True,
    bottom: bool = True,
    left: bool = True,
    right: bool = True,
    horizontal_char: str = '━',
    vertical_char: str = '┃',
    top_left_char: str = '┏',
    top_right_char: str = '┓',
    bottom_left_char: str = '┗',
    bottom_right_char: str = '┛',
    left_margin_columns: int = 0,
    right_margin_columns: int = 0,
) -> HSplit:
    """Wrap any prompt_toolkit container in a border."""

    top_border_items: List[AnyContainer] = []
    if left:
        top_border_items.append(
            Window(width=1, height=1, char=top_left_char, style=border_style)
        )

    title_text = None
    if title:
        if isinstance(title, str):
            title_text = FormattedTextControl(
                [('', f'{horizontal_char}{horizontal_char} {title} ')]
            )
        else:
            title_text = FormattedTextControl(title)

    top_border_items.append(
        Window(
            title_text,
            char=horizontal_char,
            # Expand width to max available space
            dont_extend_width=False,
            style=border_style,
        )
    )
    if right:
        top_border_items.append(
            Window(width=1, height=1, char=top_right_char, style=border_style)
        )

    content_items: List[AnyContainer] = []
    if left:
        content_items.append(
            Window(
                width=1,
                height=content_height,
                char=vertical_char,
                style=border_style,
            )
        )

    if left_margin_columns > 0:
        content_items.append(
            Window(
                width=left_margin_columns,
                height=content_height,
                char=' ',
                style=border_style,
            )
        )
    content_items.append(content)
    if right_margin_columns > 0:
        content_items.append(
            Window(
                width=right_margin_columns,
                height=content_height,
                char=' ',
                style=border_style,
            )
        )

    if right:
        content_items.append(
            Window(width=1, height=2, char=vertical_char, style=border_style)
        )

    bottom_border_items: List[AnyContainer] = []
    if left:
        bottom_border_items.append(
            Window(width=1, height=1, char=bottom_left_char)
        )
    bottom_border_items.append(
        Window(
            char=horizontal_char,
            # Expand width to max available space
            dont_extend_width=False,
        )
    )
    if right:
        bottom_border_items.append(
            Window(width=1, height=1, char=bottom_right_char)
        )

    rows: List[AnyContainer] = []
    if top:
        rows.append(VSplit(top_border_items, height=1, padding=0))
    rows.append(VSplit(content_items, height=content_height))
    if bottom:
        rows.append(
            VSplit(bottom_border_items, height=1, padding=0, style=border_style)
        )

    return HSplit(rows, style=base_style)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Serializes an Environment into a JSON file."""

from __future__ import print_function

import ntpath
import os
import posixpath
import re

# Disable super() warnings since this file must be Python 2 compatible.
# pylint: disable=super-with-arguments


class GNIVisitor(object):  # pylint: disable=useless-object-inheritance
    """Serializes portions of an Environment into a gni file.

    Example gni file:

    declare_args() {
      pw_env_setup_CIPD_DEFAULT = "//<ENVIRONMENT_DIR>/cipd/packages/default"
      pw_env_setup_CIPD_PIGWEED = "//<ENVIRONMENT_DIR>/cipd/packages/pigweed"
      pw_env_setup_CIPD_ARM = "//<ENVIRONMENT_DIR>/cipd/packages/arm"
      pw_env_setup_CIPD_PYTHON = "//<ENVIRONMENT_DIR>/cipd/packages/python"
      pw_env_setup_CIPD_BAZEL = "//<ENVIRONMENT_DIR>/cipd/packages/bazel"
      pw_env_setup_CIPD_LUCI = "//<ENVIRONMENT_DIR>/cipd/packages/luci"
      pw_env_setup_VIRTUAL_ENV = "//<ENVIRONMENT_DIR>/pigweed-venv"
    }
    """

    def __init__(self, project_root, *args, **kwargs):
        super(GNIVisitor, self).__init__(*args, **kwargs)
        self._project_root = project_root
        self._lines = []

    def serialize(self, env, outs):
        self._lines.append(
            """
# This file is automatically generated by Pigweed's environment setup. Do not
# edit it manually or check it in.
""".strip()
        )

        self._lines.append('declare_args() {')

        env.accept(self)

        self._lines.append('}')

        for line in self._lines:
            print(line, file=outs)
        self._lines = []

    def _abspath_to_gn_path(self, path):
        gn_path = os.path.relpath(path, start=self._project_root)
        if os.name == 'nt':
            # GN paths are posix-style, so convert to posix. This
            # find-and-replace is a little crude, but python 2.7 doesn't support
            # pathlib.
            gn_path = gn_path.replace(ntpath.sep, posixpath.sep)
        return '//{}'.format(gn_path)

    def visit_set(self, set):  # pylint: disable=redefined-builtin
        match = re.search(r'PW_(.*)_CIPD_INSTALL_DIR', set.name)
        if match:
            name = 'pw_env_setup_CIPD_{}'.format(match.group(1))
            self._lines.append(
                '  {} = "{}"'.format(name, self._abspath_to_gn_path(set.value))
            )

        if set.name == 'VIRTUAL_ENV':
            self._lines.append(
                '  pw_env_setup_VIRTUAL_ENV = "{}"'.format(
                    self._abspath_to_gn_path(set.value)
                )
            )

        if set.name == 'PW_PACKAGE_ROOT':
            self._lines.append(
                '  pw_env_setup_PACKAGE_ROOT = "{}"'.format(
                    self._abspath_to_gn_path(set.value)
                )
            )

    def visit_clear(self, clear):
        pass

    def visit_remove(self, remove):
        pass

    def visit_prepend(self, prepend):
        pass

    def visit_append(self, append):
        pass

    def visit_echo(self, echo):
        pass

    def visit_comment(self, comment):
        pass

    def visit_command(self, command):
        pass

    def visit_doctor(self, doctor):
        pass

    def visit_blank_line(self, blank_line):
        pass

    def visit_function(self, function):
        pass

    def visit_hash(self, hash):  # pylint: disable=redefined-builtin
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/benchmark.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_rpc/benchmark.proto',
  package='pw.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16pw_rpc/benchmark.proto\x12\x06pw.rpc\"\x1a\n\x07Payload\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x32u\n\tBenchmark\x12-\n\tUnaryEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload\x12\x39\n\x11\x42idirectionalEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload(\x01\x30\x01\x62\x06proto3'
)




_PAYLOAD = _descriptor.Descriptor(
  name='Payload',
  full_name='pw.rpc.Payload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='payload', full_name='pw.rpc.Payload.payload', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=60,
)

DESCRIPTOR.message_types_by_name['Payload'] = _PAYLOAD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Payload = _reflection.GeneratedProtocolMessageType('Payload', (_message.Message,), {
  'DESCRIPTOR' : _PAYLOAD,
  '__module__' : 'pw_rpc.benchmark_pb2'
  # @@protoc_insertion_point(class_scope:pw.rpc.Payload)
  })
_sym_db.RegisterMessage(Payload)



_BENCHMARK = _descriptor.ServiceDescriptor(
  name='Benchmark',
  full_name='pw.rpc.Benchmark',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=62,
  serialized_end=179,
  methods=[
  _descriptor.MethodDescriptor(
    name='UnaryEcho',
    full_name='pw.rpc.Benchmark.UnaryEcho',
    index=0,
    containing_service=None,
    input_type=_PAYLOAD,
    output_type=_PAYLOAD,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BidirectionalEcho',
    full_name='pw.rpc.Benchmark.BidirectionalEcho',
    index=1,
    containing_service=None,
    input_type=_PAYLOAD,
    output_type=_PAYLOAD,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BENCHMARK)

DESCRIPTOR.services_by_name['Benchmark'] = _BENCHMARK

# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""The callback-based pw_rpc client implementation."""

import inspect
import logging
import textwrap
from typing import Any, Callable, Dict, Iterable, Optional, Type

from pw_status import Status
from google.protobuf.message import Message

from pw_rpc import client, descriptors
from pw_rpc.client import PendingRpc, PendingRpcs
from pw_rpc.descriptors import Channel, Method, Service

from pw_rpc.callback_client.call import (
    UseDefault,
    OptionalTimeout,
    CallTypeT,
    UnaryResponse,
    StreamResponse,
    Call,
    UnaryCall,
    ServerStreamingCall,
    ClientStreamingCall,
    BidirectionalStreamingCall,
    OnNextCallback,
    OnCompletedCallback,
    OnErrorCallback,
)

_LOG = logging.getLogger(__package__)


class _MethodClient:
    """A method that can be invoked for a particular channel."""

    def __init__(
        self,
        client_impl: 'Impl',
        rpcs: PendingRpcs,
        channel: Channel,
        method: Method,
        default_timeout_s: Optional[float],
    ) -> None:
        self._impl = client_impl
        self._rpcs = rpcs
        self._rpc = PendingRpc(channel, method.service, method)
        self.default_timeout_s: Optional[float] = default_timeout_s

    @property
    def channel(self) -> Channel:
        return self._rpc.channel

    @property
    def method(self) -> Method:
        return self._rpc.method

    @property
    def service(self) -> Service:
        return self._rpc.service

    @property
    def request(self) -> type:
        """Returns the request proto class."""
        return self.method.request_type

    @property
    def response(self) -> type:
        """Returns the response proto class."""
        return self.method.response_type

    def __repr__(self) -> str:
        return self.help()

    def help(self) -> str:
        """Returns a help message about this RPC."""
        function_call = self.method.full_name + '('

        docstring = inspect.getdoc(self.__call__)  # type: ignore[operator] # pylint: disable=no-member
        assert docstring is not None

        annotation = inspect.Signature.from_callable(self).return_annotation  # type: ignore[arg-type] # pylint: disable=line-too-long
        if isinstance(annotation, type):
            annotation = annotation.__name__

        arg_sep = f',\n{" " * len(function_call)}'
        return (
            f'{function_call}'
            f'{arg_sep.join(descriptors.field_help(self.method.request_type))})'
            f'\n\n{textwrap.indent(docstring, "  ")}\n\n'
            f'  Returns {annotation}.'
        )

    def _start_call(
        self,
        call_type: Type[CallTypeT],
        request: Optional[Message],
        timeout_s: OptionalTimeout,
        on_next: Optional[OnNextCallback],
        on_completed: Optional[OnCompletedCallback],
        on_error: Optional[OnErrorCallback],
        ignore_errors: bool = False,
    ) -> CallTypeT:
        """Creates the Call object and invokes the RPC using it."""
        if timeout_s is UseDefault.VALUE:
            timeout_s = self.default_timeout_s

        call = call_type(
            self._rpcs, self._rpc, timeout_s, on_next, on_completed, on_error
        )
        call._invoke(request, ignore_errors)  # pylint: disable=protected-access
        return call

    def _client_streaming_call_type(
        self, base: Type[CallTypeT]
    ) -> Type[CallTypeT]:
        """Creates a client or bidirectional stream call type.

        Applies the signature from the request protobuf to the send method.
        """

        def send(
            self, _rpc_request_proto: Optional[Message] = None, **request_fields
        ) -> None:
            ClientStreamingCall.send(self, _rpc_request_proto, **request_fields)

        _apply_protobuf_signature(self.method, send)

        return type(
            f'{self.method.name}_{base.__name__}', (base,), dict(send=send)
        )


def _function_docstring(method: Method) -> str:
    return f'''\
Invokes the {method.full_name} {method.type.sentence_name()} RPC.

This function accepts either the request protobuf fields as keyword arguments or
a request protobuf as a positional argument.
'''


def _update_call_method(method: Method, function: Callable) -> None:
    """Updates the name, docstring, and parameters to match a method."""
    function.__name__ = method.full_name
    function.__doc__ = _function_docstring(method)
    _apply_protobuf_signature(method, function)


def _apply_protobuf_signature(method: Method, function: Callable) -> None:
    """Update a function signature to accept proto arguments.

    In order to have good tab completion and help messages, update the function
    signature to accept only keyword arguments for the proto message fields.
    This doesn't actually change the function signature -- it just updates how
    it appears when inspected.
    """
    sig = inspect.signature(function)

    params = [next(iter(sig.parameters.values()))]  # Get the "self" parameter
    params += method.request_parameters()
    params.append(
        inspect.Parameter('pw_rpc_timeout_s', inspect.Parameter.KEYWORD_ONLY)
    )

    function.__signature__ = sig.replace(  # type: ignore[attr-defined]
        parameters=params
    )


class _UnaryMethodClient(_MethodClient):
    def invoke(
        self,
        request: Optional[Message] = None,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        request_args: Optional[Dict[str, Any]] = None,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> UnaryCall:
        """Invokes the unary RPC and returns a call object."""
        return self._start_call(
            UnaryCall,
            self.method.get_request(request, request_args),
            timeout_s,
            on_next,
            on_completed,
            on_error,
        )

    def open(
        self,
        request: Optional[Message] = None,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        request_args: Optional[Dict[str, Any]] = None,
    ) -> UnaryCall:
        """Invokes the unary RPC and returns a call object."""
        return self._start_call(
            UnaryCall,
            self.method.get_request(request, request_args),
            None,
            on_next,
            on_completed,
            on_error,
            True,
        )


class _ServerStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        request: Optional[Message] = None,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        request_args: Optional[Dict[str, Any]] = None,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> ServerStreamingCall:
        """Invokes the server streaming RPC and returns a call object."""
        return self._start_call(
            ServerStreamingCall,
            self.method.get_request(request, request_args),
            timeout_s,
            on_next,
            on_completed,
            on_error,
        )

    def open(
        self,
        request: Optional[Message] = None,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        request_args: Optional[Dict[str, Any]] = None,
    ) -> ServerStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._start_call(
            ServerStreamingCall,
            self.method.get_request(request, request_args),
            None,
            on_next,
            on_completed,
            on_error,
            True,
        )


class _ClientStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> ClientStreamingCall:
        """Invokes the client streaming RPC and returns a call object"""
        return self._start_call(
            self._client_streaming_call_type(ClientStreamingCall),
            None,
            timeout_s,
            on_next,
            on_completed,
            on_error,
            True,
        )

    def open(
        self,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
    ) -> ClientStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._start_call(
            self._client_streaming_call_type(ClientStreamingCall),
            None,
            None,
            on_next,
            on_completed,
            on_error,
            True,
        )

    def __call__(
        self,
        requests: Iterable[Message] = (),
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> UnaryResponse:
        return self.invoke().finish_and_wait(requests, timeout_s=timeout_s)


class _BidirectionalStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> BidirectionalStreamingCall:
        """Invokes the bidirectional streaming RPC and returns a call object."""
        return self._start_call(
            self._client_streaming_call_type(BidirectionalStreamingCall),
            None,
            timeout_s,
            on_next,
            on_completed,
            on_error,
        )

    def open(
        self,
        on_next: Optional[OnNextCallback] = None,
        on_completed: Optional[OnCompletedCallback] = None,
        on_error: Optional[OnErrorCallback] = None,
    ) -> BidirectionalStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._start_call(
            self._client_streaming_call_type(BidirectionalStreamingCall),
            None,
            None,
            on_next,
            on_completed,
            on_error,
            True,
        )

    def __call__(
        self,
        requests: Iterable[Message] = (),
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> StreamResponse:
        return self.invoke().finish_and_wait(requests, timeout_s=timeout_s)


def _method_client_docstring(method: Method) -> str:
    return f'''\
Class that invokes the {method.full_name} {method.type.sentence_name()} RPC.

Calling this directly invokes the RPC synchronously. The RPC can be invoked
asynchronously using the invoke method.
'''


class Impl(client.ClientImpl):
    """Callback-based ClientImpl, for use with pw_rpc.Client."""

    def __init__(
        self,
        default_unary_timeout_s: Optional[float] = None,
        default_stream_timeout_s: Optional[float] = None,
        cancel_duplicate_calls: Optional[bool] = True,
    ) -> None:
        super().__init__()
        self._default_unary_timeout_s = default_unary_timeout_s
        self._default_stream_timeout_s = default_stream_timeout_s

        # Temporary workaround for clients that rely on mulitple in-flight
        # instances of an RPC on the same channel, which is not supported.
        # TODO(hepler): Remove this option when clients have updated.
        self._cancel_duplicate_calls = cancel_duplicate_calls

    @property
    def default_unary_timeout_s(self) -> Optional[float]:
        return self._default_unary_timeout_s

    @property
    def default_stream_timeout_s(self) -> Optional[float]:
        return self._default_stream_timeout_s

    def method_client(self, channel: Channel, method: Method) -> _MethodClient:
        """Returns an object that invokes a method using the given chanel."""

        # Temporarily attach the cancel_duplicate_calls option to the
        # PendingRpcs object.
        # TODO(hepler): Remove this workaround.
        assert self.rpcs
        self.rpcs.cancel_duplicate_calls = (  # type: ignore[attr-defined]
            self._cancel_duplicate_calls
        )

        if method.type is Method.Type.UNARY:
            return self._create_unary_method_client(
                channel, method, self.default_unary_timeout_s
            )

        if method.type is Method.Type.SERVER_STREAMING:
            return self._create_server_streaming_method_client(
                channel, method, self.default_stream_timeout_s
            )

        if method.type is Method.Type.CLIENT_STREAMING:
            return self._create_method_client(
                _ClientStreamingMethodClient,
                channel,
                method,
                self.default_unary_timeout_s,
            )

        if method.type is Method.Type.BIDIRECTIONAL_STREAMING:
            return self._create_method_client(
                _BidirectionalStreamingMethodClient,
                channel,
                method,
                self.default_stream_timeout_s,
            )

        raise AssertionError(f'Unknown method type {method.type}')

    def _create_method_client(
        self,
        base: type,
        channel: Channel,
        method: Method,
        default_timeout_s: Optional[float],
        **fields,
    ):
        """Creates a _MethodClient derived class customized for the method."""
        method_client_type = type(
            f'{method.name}{base.__name__}',
            (base,),
            dict(__doc__=_method_client_docstring(method), **fields),
        )
        return method_client_type(
            self, self.rpcs, channel, method, default_timeout_s
        )

    def _create_unary_method_client(
        self,
        channel: Channel,
        method: Method,
        default_timeout_s: Optional[float],
    ) -> _UnaryMethodClient:
        """Creates a _UnaryMethodClient with a customized __call__ method."""

        # TODO(hepler): Use / to mark the first arg as positional-only
        #     when when Python 3.7 support is no longer required.
        def call(
            self: _UnaryMethodClient,
            _rpc_request_proto: Optional[Message] = None,
            *,
            pw_rpc_timeout_s: OptionalTimeout = UseDefault.VALUE,
            **request_fields,
        ) -> UnaryResponse:
            return self.invoke(
                self.method.get_request(_rpc_request_proto, request_fields)
            ).wait(pw_rpc_timeout_s)

        _update_call_method(method, call)
        return self._create_method_client(
            _UnaryMethodClient,
            channel,
            method,
            default_timeout_s,
            __call__=call,
        )

    def _create_server_streaming_method_client(
        self,
        channel: Channel,
        method: Method,
        default_timeout_s: Optional[float],
    ) -> _ServerStreamingMethodClient:
        """Creates _ServerStreamingMethodClient with custom __call__ method."""

        # TODO(hepler): Use / to mark the first arg as positional-only
        #     when when Python 3.7 support is no longer required.
        def call(
            self: _ServerStreamingMethodClient,
            _rpc_request_proto: Optional[Message] = None,
            *,
            pw_rpc_timeout_s: OptionalTimeout = UseDefault.VALUE,
            **request_fields,
        ) -> StreamResponse:
            return self.invoke(
                self.method.get_request(_rpc_request_proto, request_fields)
            ).wait(pw_rpc_timeout_s)

        _update_call_method(method, call)
        return self._create_method_client(
            _ServerStreamingMethodClient,
            channel,
            method,
            default_timeout_s,
            __call__=call,
        )

    def handle_response(
        self,
        rpc: PendingRpc,
        context: Call,
        payload,
        *,
        args: tuple = (),
        kwargs: Optional[dict] = None,
    ) -> None:
        """Invokes the callback associated with this RPC."""
        assert not args and not kwargs, 'Forwarding args & kwargs not supported'
        context._handle_response(payload)  # pylint: disable=protected-access

    def handle_completion(
        self,
        rpc: PendingRpc,
        context: Call,
        status: Status,
        *,
        args: tuple = (),
        kwargs: Optional[dict] = None,
    ):
        assert not args and not kwargs, 'Forwarding args & kwargs not supported'
        context._handle_completion(status)  # pylint: disable=protected-access

    def handle_error(
        self,
        rpc: PendingRpc,
        context: Call,
        status: Status,
        *,
        args: tuple = (),
        kwargs: Optional[dict] = None,
    ) -> None:
        assert not args and not kwargs, 'Forwarding args & kwargs not supported'
        context._handle_error(status)  # pylint: disable=protected-access



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Window pane base class."""

import asyncio
import logging
from threading import Thread
import time
from typing import Callable, Optional

from pw_console.get_pw_console_app import get_pw_console_app


class PluginMixin:
    """Handles background task management in a Pigweed Console plugin.

    Pigweed Console plugins can inherit from this class if they require running
    tasks in the background. This is important as any plugin code not in its
    own dedicated thread can potentially block the user interface

    Example usage: ::

        import logging
        from pw_console.plugin_mixin import PluginMixin
        from pw_console.widgets import WindowPaneToolbar

        class AwesomeToolbar(WindowPaneToolbar, PluginMixin):
            TOOLBAR_HEIGHT = 1

            def __init__(self, *args, **kwargs):
                # Call parent class WindowPaneToolbar.__init__
                super().__init__(*args, **kwargs)

                # Set PluginMixin to execute
                # self._awesome_background_task every 10 seconds.
                self.plugin_init(
                    plugin_callback=self._awesome_background_task,
                    plugin_callback_frequency=10.0,
                    plugin_logger_name='awesome_toolbar_plugin')

            # This function will be run in a separate thread every 10 seconds.
            def _awesome_background_task(self) -> bool:
                time.sleep(1)  # Do real work here.

                if self.new_data_processed:
                    # If new data was processed, and the user interface
                    # should be updated return True.

                    # Log using self.plugin_logger for debugging.
                    self.plugin_logger.debug('New data processed')

                    # Return True to signal a UI redraw.
                    return True

                # Returning False means no updates needed.
                return False

    Attributes:
        plugin_callback: Callable that is run in a background thread.
        plugin_callback_frequency: Number of seconds to wait between
            executing plugin_callback.
        plugin_logger: logging instance for this plugin. Useful for debugging
            code running in a separate thread.
        plugin_callback_future: `Future`_ object for the plugin background task.
        plugin_event_loop: asyncio event loop running in the background thread.
        plugin_enable_background_task: If True, keep periodically running
            plugin_callback at the desired frequency. If False the background
            task will stop.

    .. _Future: https://docs.python.org/3/library/asyncio-future.html
    """

    def plugin_init(
        self,
        plugin_callback: Optional[Callable[..., bool]] = None,
        plugin_callback_frequency: float = 30.0,
        plugin_logger_name: Optional[str] = 'pw_console_plugins',
    ) -> None:
        """Call this on __init__() to set plugin background task variables.

        Args:
            plugin_callback: Callable to run in a separate thread from the
                Pigweed Console UI. This function should return True if the UI
                should be redrawn after execution.
            plugin_callback_frequency: Number of seconds to wait between
                executing plugin_callback.
            plugin_logger_name: Unique name for this plugin's Python
                logger. Useful for debugging code running in a separate thread.
        """
        self.plugin_callback = plugin_callback
        self.plugin_callback_frequency = plugin_callback_frequency
        self.plugin_logger = logging.getLogger(plugin_logger_name)

        self.plugin_callback_future = None

        # Event loop for executing plugin code.
        self.plugin_event_loop = asyncio.new_event_loop()
        self.plugin_enable_background_task = True

    def plugin_start(self):
        """Function used to start this plugin's background thead and task."""

        # Create an entry point for the plugin thread.
        def _plugin_thread_entry():
            # Disable log propagation
            self.plugin_logger.propagate = False
            asyncio.set_event_loop(self.plugin_event_loop)
            self.plugin_event_loop.run_forever()

        # Create a thread for running user code so the UI isn't blocked.
        thread = Thread(target=_plugin_thread_entry, args=(), daemon=True)
        thread.start()

        self.plugin_logger.debug('Starting plugin: %s', self)
        if self.plugin_callback is None:
            return

        self.plugin_enable_background_task = True
        self.plugin_callback_future = asyncio.run_coroutine_threadsafe(
            # This function will be executed in a separate thread.
            self._plugin_periodically_run_callback(),
            # Using this asyncio event loop.
            self.plugin_event_loop,
        )  # type: ignore

    def plugin_stop(self):
        self.plugin_enable_background_task = False

    async def _plugin_periodically_run_callback(self) -> None:
        while self.plugin_enable_background_task:
            start_time = time.time()
            # Run the callback and redraw the UI if return value is True
            if self.plugin_callback and self.plugin_callback():
                get_pw_console_app().redraw_ui()
            run_time = time.time() - start_time
            await asyncio.sleep(self.plugin_callback_frequency - run_time)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class ThreadState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Enum:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ThreadState._Enum.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: ThreadState._Enum.ValueType  # 0
        """Thread state is invalid or cannot be expressed by this enum."""
        INTERRUPT_HANDLER: ThreadState._Enum.ValueType  # 1
        """Interrupt handling is often done on a stack that isn't associated with a
        true RTOS thread. This state indicates the provided thread info is for an
        interrupt handler.
        """
        RUNNING: ThreadState._Enum.ValueType  # 2
        """This is the currently active thread as marked by the RTOS. In crashes in
        interrupt contexts, this isn’t necessarily the thread that crashed.
        """
        READY: ThreadState._Enum.ValueType  # 3
        """Thread is ready to run, but isn’t currently running."""
        SUSPENDED: ThreadState._Enum.ValueType  # 4
        """The thread is not ready to run, and will not be ready to run until it is
        explicitly resumed.
        """
        BLOCKED: ThreadState._Enum.ValueType  # 5
        """The thread is waiting on something before it can run again."""
        INACTIVE: ThreadState._Enum.ValueType  # 6
        """The thread is either not yet initialized, or has terminated. In other
        words, this thread is a suspended thread that cannot be unsuspended.
        """

    class Enum(_Enum, metaclass=_EnumEnumTypeWrapper): ...
    UNKNOWN: ThreadState.Enum.ValueType  # 0
    """Thread state is invalid or cannot be expressed by this enum."""
    INTERRUPT_HANDLER: ThreadState.Enum.ValueType  # 1
    """Interrupt handling is often done on a stack that isn't associated with a
    true RTOS thread. This state indicates the provided thread info is for an
    interrupt handler.
    """
    RUNNING: ThreadState.Enum.ValueType  # 2
    """This is the currently active thread as marked by the RTOS. In crashes in
    interrupt contexts, this isn’t necessarily the thread that crashed.
    """
    READY: ThreadState.Enum.ValueType  # 3
    """Thread is ready to run, but isn’t currently running."""
    SUSPENDED: ThreadState.Enum.ValueType  # 4
    """The thread is not ready to run, and will not be ready to run until it is
    explicitly resumed.
    """
    BLOCKED: ThreadState.Enum.ValueType  # 5
    """The thread is waiting on something before it can run again."""
    INACTIVE: ThreadState.Enum.ValueType  # 6
    """The thread is either not yet initialized, or has terminated. In other
    words, this thread is a suspended thread that cannot be unsuspended.
    """

    def __init__(
        self,
    ) -> None: ...

global___ThreadState = ThreadState

class Thread(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    ACTIVE_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    RAW_BACKTRACE_FIELD_NUMBER: builtins.int
    SYMBOLIZED_BACKTRACE_FIELD_NUMBER: builtins.int
    RAW_STACK_FIELD_NUMBER: builtins.int
    STACK_START_POINTER_FIELD_NUMBER: builtins.int
    STACK_END_POINTER_FIELD_NUMBER: builtins.int
    STACK_POINTER_FIELD_NUMBER: builtins.int
    CPU_USAGE_HUNDREDTHS_FIELD_NUMBER: builtins.int
    STACK_POINTER_EST_PEAK_FIELD_NUMBER: builtins.int
    name: builtins.bytes
    """Thread names must be unique; this allows extensions of Snapshot to augment
    threads with additional data. This should either be human readable text, or
    tokenized data (e.g. base-64 encoded or binary data).
    """
    active: builtins.bool
    """This field has been deprecatdin favor of using the state enum to report
    RUNNING or INTERRUPT_CONTEXT to mark them as active.

    Whether or not this thread is the thread is the currently active context
    at the time of capture. For multi-thread dumps, this field should only be
    set on ONE thread.
    """
    state: global___ThreadState.Enum.ValueType
    """A summarized thread state. RTOS-specific extensions of the Thread message
    may provide more specific thread state information.
    """
    @property
    def raw_backtrace(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Contents of a stack trace. It is expected that this stack is pre-walked,
        and contains addresses. Most recent stack events are at the beginning of
        the captured stack trace.
        """
    @property
    def symbolized_backtrace(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Results of symbolizing stack_entries. This is usually not provided by the
        device, but instead by server/host side processing.
        """
    raw_stack: builtins.bytes
    """This should contain the raw contents of the thread's stack. This might not
    match stack_size. It can be larger due to a stack overflow, or smaller due
    to the implementation deciding to only capture a portion of the stack.
    Partial stack captures are typically a result of storage/memory
    limitations.
    """
    stack_start_pointer: builtins.int
    """The address this thread's stack pointer began at. For descending stacks,
    this is the highest address of the stack bounds. For ascending stacks, this
    is the lowest address of the stack bounds.
    """
    stack_end_pointer: builtins.int
    """The furthest permitted address from where this thread's stack pointer
    began. For descending stacks, this is the lowest address of the stack
    bounds. For ascending stacks, this is the highest address of the stack
    bounds.
    """
    stack_pointer: builtins.int
    """The current stack pointer of this thread."""
    cpu_usage_hundredths: builtins.int
    """CPU usage info. This is the percentage of CPU time the thread has been
    active in hundredths of a percent. (e.g. 5.00% = 500u)
    """
    stack_pointer_est_peak: builtins.int
    """The address of highest estimated currently used in the thread stack.
    Percentage of bytes used can be calculated by:
    (stack_estimate_max_addr-stack_start_pointer) /
    (stack_end_pointer-stack_start_pointer) * 100%
    """
    def __init__(
        self,
        *,
        name: builtins.bytes = ...,
        active: builtins.bool = ...,
        state: global___ThreadState.Enum.ValueType = ...,
        raw_backtrace: collections.abc.Iterable[builtins.int] | None = ...,
        symbolized_backtrace: collections.abc.Iterable[builtins.str] | None = ...,
        raw_stack: builtins.bytes = ...,
        stack_start_pointer: builtins.int | None = ...,
        stack_end_pointer: builtins.int | None = ...,
        stack_pointer: builtins.int | None = ...,
        cpu_usage_hundredths: builtins.int | None = ...,
        stack_pointer_est_peak: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_cpu_usage_hundredths", b"_cpu_usage_hundredths", "_stack_end_pointer", b"_stack_end_pointer", "_stack_pointer", b"_stack_pointer", "_stack_pointer_est_peak", b"_stack_pointer_est_peak", "_stack_start_pointer", b"_stack_start_pointer", "cpu_usage_hundredths", b"cpu_usage_hundredths", "stack_end_pointer", b"stack_end_pointer", "stack_pointer", b"stack_pointer", "stack_pointer_est_peak", b"stack_pointer_est_peak", "stack_start_pointer", b"stack_start_pointer"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_cpu_usage_hundredths", b"_cpu_usage_hundredths", "_stack_end_pointer", b"_stack_end_pointer", "_stack_pointer", b"_stack_pointer", "_stack_pointer_est_peak", b"_stack_pointer_est_peak", "_stack_start_pointer", b"_stack_start_pointer", "active", b"active", "cpu_usage_hundredths", b"cpu_usage_hundredths", "name", b"name", "raw_backtrace", b"raw_backtrace", "raw_stack", b"raw_stack", "stack_end_pointer", b"stack_end_pointer", "stack_pointer", b"stack_pointer", "stack_pointer_est_peak", b"stack_pointer_est_peak", "stack_start_pointer", b"stack_start_pointer", "state", b"state", "symbolized_backtrace", b"symbolized_backtrace"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_cpu_usage_hundredths", b"_cpu_usage_hundredths"]) -> typing_extensions.Literal["cpu_usage_hundredths"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_stack_end_pointer", b"_stack_end_pointer"]) -> typing_extensions.Literal["stack_end_pointer"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_stack_pointer", b"_stack_pointer"]) -> typing_extensions.Literal["stack_pointer"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_stack_pointer_est_peak", b"_stack_pointer_est_peak"]) -> typing_extensions.Literal["stack_pointer_est_peak"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_stack_start_pointer", b"_stack_start_pointer"]) -> typing_extensions.Literal["stack_start_pointer"] | None: ...

global___Thread = Thread

class SnapshotThreadInfo(google.protobuf.message.Message):
    """This message overlays the pw.snapshot.Snapshot proto. It's valid to encode
    this message to the same sink that a Snapshot proto is being written to.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    THREADS_FIELD_NUMBER: builtins.int
    @property
    def threads(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Thread]: ...
    def __init__(
        self,
        *,
        threads: collections.abc.Iterable[global___Thread] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["threads", b"threads"]) -> None: ...

global___SnapshotThreadInfo = SnapshotThreadInfo



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console entry point."""

import argparse
import inspect
import logging
from pathlib import Path
import sys
from typing import Optional, Dict

from pw_cli import log as pw_cli_log
from pw_cli import argument_types

from pw_console import PwConsoleEmbed
from pw_console.python_logging import create_temp_log_file
from pw_console.log_store import LogStore
from pw_console.plugins.calc_pane import CalcPane
from pw_console.plugins.clock_pane import ClockPane
from pw_console.plugins.twenty48_pane import Twenty48Pane
from pw_console.test_mode import FAKE_DEVICE_LOGGER_NAME

_LOG = logging.getLogger(__package__)
_ROOT_LOG = logging.getLogger('')


def _build_argument_parser() -> argparse.ArgumentParser:
    """Setup argparse."""
    parser = argparse.ArgumentParser(
        prog="python -m pw_console", description=__doc__
    )

    parser.add_argument(
        '-l',
        '--loglevel',
        type=argument_types.log_level,
        default=logging.DEBUG,
        help='Set the log level' '(debug, info, warning, error, critical)',
    )

    parser.add_argument('--logfile', help='Pigweed Console log file.')

    parser.add_argument(
        '--test-mode',
        action='store_true',
        help='Enable fake log messages for testing purposes.',
    )
    parser.add_argument(
        '--config-file',
        type=Path,
        help='Path to a pw_console yaml config file.',
    )
    parser.add_argument(
        '--console-debug-log-file',
        help='Log file to send console debug messages to.',
    )

    return parser


def main() -> int:
    """Pigweed Console."""

    parser = _build_argument_parser()
    args = parser.parse_args()

    if not args.logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        args.logfile = create_temp_log_file()

    pw_cli_log.install(
        level=args.loglevel,
        use_color=True,
        hide_timestamp=False,
        log_file=args.logfile,
    )

    if args.console_debug_log_file:
        pw_cli_log.install(
            level=logging.DEBUG,
            use_color=True,
            hide_timestamp=False,
            log_file=args.console_debug_log_file,
            logger=logging.getLogger('pw_console'),
        )

    global_vars = None
    default_loggers = {}
    if args.test_mode:
        root_log_store = LogStore()
        _ROOT_LOG.addHandler(root_log_store)
        _ROOT_LOG.debug('pw_console test-mode starting...')

        fake_logger = logging.getLogger(FAKE_DEVICE_LOGGER_NAME)
        default_loggers = {
            # Don't include pw_console package logs (_LOG) in the log pane UI.
            # Add the fake logger for test_mode.
            'Fake Device': [fake_logger],
            'PwConsole Debug': [logging.getLogger('pw_console')],
            'All Logs': root_log_store,
        }
        # Give access to adding log messages from the repl via: `LOG.warning()`
        global_vars = dict(LOG=fake_logger)

    help_text = None
    app_title = None
    if args.test_mode:
        app_title = 'Console Test Mode'
        help_text = inspect.cleandoc(
            """
            Welcome to the Pigweed Console Test Mode!

            Example commands:

              rpcs.pw.rpc.EchoService.Echo(msg='hello!')

              LOG.warning('Message appears console log window.')
        """
        )

    console = PwConsoleEmbed(
        global_vars=global_vars,
        loggers=default_loggers,
        test_mode=args.test_mode,
        help_text=help_text,
        app_title=app_title,
        config_file_path=args.config_file,
    )

    overriden_window_config: Optional[Dict] = None
    # Add example plugins and log panes used to validate behavior in the Pigweed
    # Console manual test procedure: https://pigweed.dev/pw_console/testing.html
    if args.test_mode:
        fake_logger.propagate = False
        console.setup_python_logging(loggers_with_no_propagation=[fake_logger])

        _ROOT_LOG.debug('pw_console.PwConsoleEmbed init complete')
        _ROOT_LOG.debug('Adding plugins...')
        console.add_window_plugin(ClockPane())
        console.add_window_plugin(CalcPane())
        console.add_floating_window_plugin(
            Twenty48Pane(include_resize_handle=False), left=4
        )
        _ROOT_LOG.debug('Starting prompt_toolkit full-screen application...')

        overriden_window_config = {
            'Split 1 stacked': {
                'Fake Device': None,
                'Fake Keys': {
                    'duplicate_of': 'Fake Device',
                    'filters': {
                        'keys': {'regex': '[^ ]+'},
                    },
                },
                'Fake USB': {
                    'duplicate_of': 'Fake Device',
                    'filters': {
                        'module': {'regex': 'USB'},
                    },
                },
            },
            'Split 2 tabbed': {
                'Python Repl': None,
                'All Logs': None,
                'PwConsole Debug': None,
                'Calculator': None,
                'Clock': None,
            },
        }

    console.embed(override_window_config=overriden_window_config)

    if args.logfile:
        print(f'Logs saved to: {args.logfile}')

    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/internal/packet.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_rpc/internal/packet.proto',
  package='pw.rpc.internal',
  syntax='proto3',
  serialized_options=b'\n\033dev.pigweed.pw_rpc.internal',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cpw_rpc/internal/packet.proto\x12\x0fpw.rpc.internal\"\xa3\x01\n\tRpcPacket\x12)\n\x04type\x18\x01 \x01(\x0e\x32\x1b.pw.rpc.internal.PacketType\x12\x12\n\nchannel_id\x18\x02 \x01(\r\x12\x12\n\nservice_id\x18\x03 \x01(\x07\x12\x11\n\tmethod_id\x18\x04 \x01(\x07\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\x12\x0e\n\x06status\x18\x06 \x01(\r\x12\x0f\n\x07\x63\x61ll_id\x18\x07 \x01(\r*\x9c\x01\n\nPacketType\x12\x0b\n\x07REQUEST\x10\x00\x12\x11\n\rCLIENT_STREAM\x10\x02\x12\x10\n\x0c\x43LIENT_ERROR\x10\x04\x12\x1d\n\x19\x43LIENT_REQUEST_COMPLETION\x10\x08\x12\x0c\n\x08RESPONSE\x10\x01\x12\x10\n\x0cSERVER_ERROR\x10\x05\x12\x11\n\rSERVER_STREAM\x10\x07\"\x04\x08\x03\x10\x03\"\x04\x08\x06\x10\x06\x42\x1d\n\x1b\x64\x65v.pigweed.pw_rpc.internalb\x06proto3'
)

_PACKETTYPE = _descriptor.EnumDescriptor(
  name='PacketType',
  full_name='pw.rpc.internal.PacketType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REQUEST', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_STREAM', index=1, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ERROR', index=2, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_REQUEST_COMPLETION', index=3, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RESPONSE', index=4, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVER_ERROR', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVER_STREAM', index=6, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=216,
  serialized_end=372,
)
_sym_db.RegisterEnumDescriptor(_PACKETTYPE)

PacketType = enum_type_wrapper.EnumTypeWrapper(_PACKETTYPE)
REQUEST = 0
CLIENT_STREAM = 2
CLIENT_ERROR = 4
CLIENT_REQUEST_COMPLETION = 8
RESPONSE = 1
SERVER_ERROR = 5
SERVER_STREAM = 7



_RPCPACKET = _descriptor.Descriptor(
  name='RpcPacket',
  full_name='pw.rpc.internal.RpcPacket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='pw.rpc.internal.RpcPacket.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_id', full_name='pw.rpc.internal.RpcPacket.channel_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_id', full_name='pw.rpc.internal.RpcPacket.service_id', index=2,
      number=3, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_id', full_name='pw.rpc.internal.RpcPacket.method_id', index=3,
      number=4, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='pw.rpc.internal.RpcPacket.payload', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='pw.rpc.internal.RpcPacket.status', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='call_id', full_name='pw.rpc.internal.RpcPacket.call_id', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=213,
)

_RPCPACKET.fields_by_name['type'].enum_type = _PACKETTYPE
DESCRIPTOR.message_types_by_name['RpcPacket'] = _RPCPACKET
DESCRIPTOR.enum_types_by_name['PacketType'] = _PACKETTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RpcPacket = _reflection.GeneratedProtocolMessageType('RpcPacket', (_message.Message,), {
  'DESCRIPTOR' : _RPCPACKET,
  '__module__' : 'pw_rpc.internal.packet_pb2'
  # @@protoc_insertion_point(class_scope:pw.rpc.internal.RpcPacket)
  })
_sym_db.RegisterMessage(RpcPacket)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
